from typing import Any


class RawFormat(object):
    model_type = Any
    formatted_type = Any
    validators = ()

    @classmethod
    def init(self):
        self.validators = []

    @classmethod
    def validate_type(self, data: formatted_type):
        if not isinstance(data, self.formatted_type):
            error = 'Expected {}, but found {}.'
            return error.format(
                self.formatted_type.__name__,
                data.__class__.__name__,
            )

    @classmethod
    def load(self, data: formatted_type) -> model_type:
        error = self.validate_type(data)
        if error:
            return None, error
        return data, None

    @staticmethod
    def dump(data: model_type) -> formatted_type:
        return data

    @classmethod
    def validate_content(self, data):
        for validator in self.validators:
            error = validator(data)
            if error:
                return error

    @classmethod
    def add_validator(self, validator):
        self.validators.append(validator)

    @classmethod
    def add_constraints(self, constraints, name=None):
        for constraint, option in constraints.items():
            if hasattr(self, constraint):
                getattr(self, constraint)(option)
            else:
                message = 'The {} format has no constraint named {}.'
                name = name or self.__name__
                raise AttributeError(message.format(name, constraint))

    @classmethod
    def map(self, name):
        """
        Instructs a schema to use an alternate name when reading during a load
        or writing during a dump.

        Arguments:
            name (str): The key the value uses in a schema's serialized state.
        """
        self.formatted_name = name

    @classmethod
    def compile(self, enabled):
        loc = {}
        context = {'id': 0}
        compile_dump = self.compile_dump('data', context)
        inline_dump = 'def dump(data):\n return {}\n'.format(compile_dump)
        exec(inline_dump, globals(), loc)
        self.dump = loc['dump']

    @staticmethod
    def compile_dump(scope, context):
        return scope



from typing import get_type_hints
from ..format_registry import FormatRegistry
from .base_format import BaseFormat
from .callable import Callable


class Object(BaseFormat):
    model_type = object
    formatted_type = dict
    schema_class = object
    type_hints = None
    child_formats = None

    @staticmethod
    def name_format(format, name):
        if hasattr(format, 'formatted_name'):
            return format

        class NamedFormat(format):
            formatted_name = name
        return NamedFormat

    @classmethod
    def init(self):
        super().init()
        type_hints = get_type_hints(self.schema_class)
        self.child_formats = [
            (name, self.name_format(FormatRegistry.get(type), name))
            for name, type in type_hints.items()
        ]
        self.callable_formats = [
            (name, format)
            for name, format in self.child_formats
            if issubclass(format, Callable)
        ]
        if self.callable_formats:
            original_dump = self.dump

            def dump(data):
                return self.dump_callables(data, original_dump(data))
            self.dump = dump
        self.child_formats = [
            (name, format)
            for name, format in self.child_formats
            if not issubclass(format, Callable)
        ]
        self.formatted_names = set([
            format.formatted_name
            for _, format in self.child_formats
        ])
        self.defaults = {
            name: getattr(self.schema_class, name)
            for name in type_hints
            if hasattr(self.schema_class, name)
        }

    @classmethod
    def validate_extra(self, data):
        return {
            key: 'This key is not expected.'
            for key in data
            if key not in self.formatted_names
        }

    @classmethod
    def validate_extra_ignore(self, data):
        return {}

    @classmethod
    def get_extra(self, data):
        return {}

    @classmethod
    def get_extra_keep(self, data):
        return {
            key: value
            for key, value in data.items()
        }

    @classmethod
    def validate_type(self, data: formatted_type):
        type_error = super().validate_type(data)
        if type_error:
            return type_error
        errors = self.validate_extra(data)
        for key, format in self.child_formats:
            data_key = format.formatted_name
            if data_key not in data and key not in self.defaults:
                error = 'This key was not found, but it is required.'
            else:
                error = None
            if error:
                errors[key] = error
        return errors

    @classmethod
    def validate_content(self, data):
        errors = {}
        for key, format in self.child_formats:
            error = format.validate_content(getattr(data, key))
            if error:
                errors[key] = error
        return errors or super().validate_content(data)

    @classmethod
    def load(self, data: formatted_type) -> model_type:
        errors = self.validate_type(data)
        if errors:
            return None, errors
        init_kwargs = self.get_extra(data)
        for key, format in self.child_formats:
            data_key = format.formatted_name
            if data_key in data:
                value, error = format.load(data[data_key])
                if error:
                    errors[key] = error
                else:
                    init_kwargs[key] = value
            else:
                init_kwargs[key] = self.defaults.get(key)
        if errors:
            return None, errors
        return self.model_type(**init_kwargs), None

    @classmethod
    def dump(self, data: model_type) -> formatted_type:
        return {
            format.formatted_name: format.dump(getattr(data, key))
            for key, format in self.child_formats
        }

    @classmethod
    def dump_callables(self, data, formatted_data):
        for key, format in self.callable_formats:
            data_key = format.formatted_name
            formatted_data[data_key] = format.dump(getattr(data, key))
        return formatted_data

    @classmethod
    def compile_dump_item(self, scope, key):
        return scope + '.' + key

    @classmethod
    def compile_dump(self, scope, context):
        values = [
            (
                '"' + format.formatted_name + '": ' +
                format.compile_dump(
                    self.compile_dump_item(scope, key),
                    context,
                )
            )
            for key, format in self.child_formats
        ] + [
            (
                '"' + format.formatted_name + '": ' +
                format.compile_dump(scope + '.' + key + '()', context)
            )
            for key, format in self.callable_formats
        ]
        return '{' + ', '.join(values) + '}'

    @classmethod
    def ignore_extra(self, enabled):
        if enabled:
            self.validate_extra = self.validate_extra_ignore
        else:
            self.validate_extra = self.validate_extra_error

    @classmethod
    def keep_extra(self, enabled):
        self.ignore_extra(enabled)
        if enabled:
            self.get_extra = self.get_extra_keep
        else:
            self.get_extra = self.get_extra_discard


Object.validate_extra_error = classmethod(Object.validate_extra.__func__)
Object.get_extra_discard = classmethod(Object.get_extra.__func__)



from .raw_format import RawFormat
from .base_format import BaseFormat
from .none_type import NoneType
from .any import Any
from .object import Object
from .dict import Dict
from .string import String
from .date_string import DateString
from .float import Float
from .integer import Integer
from .integer_string import IntegerString
from .decimal import Decimal
from .decimal_string import DecimalString
from .union import Union
from .list import List
from .callable import Callable

__all__ = [
    'RawFormat',
    'BaseFormat',
    'NoneType',
    'Any',
    'Object',
    'Dict',
    'String',
    'DateString',
    'Float',
    'Integer',
    'IntegerString',
    'Decimal',
    'DecimalString',
    'Union',
    'List',
    'Callable',
]



from .raw_format import RawFormat


class BaseFormat(RawFormat):
    @classmethod
    def load(self, data):
        error = self.validate_type(data)
        if error:
            return None, error
        return self.model_type(data), None

    @classmethod
    def dump(self, data):
        return self.formatted_type(data)

    @classmethod
    def compile_dump(self, scope, context):
        return self.formatted_type.__name__ + '(' + scope + ')'



from typing import NewType


class FormatRegistry(object):
    default_formats = {}
    default_generic_formats = {}

    @classmethod
    def set_default(self, format, type=None):
        type = type or format.model_type
        self.default_formats[type] = format

    @classmethod
    def set_generic_default(self, format, type=None):
        type = type or format.generic_type.__class__
        self.default_generic_formats[type] = format

    @classmethod
    def build_type(self, format):
        type = NewType(repr(format), format.model_type)
        type.__marsha_format__ = format
        return type

    @classmethod
    def get(self, type):
        format = getattr(type, '__marsha_format__', None)
        if format:
            return format
        format = self.default_formats.get(type, None)
        if format:
            return format
        generic_format = self.default_generic_formats.get(type.__class__, None)
        if generic_format:
            return generic_format.from_type(type)
        message = 'There is no default format for {}.'.format(type.__name__)
        raise TypeError(message)

    @classmethod
    def register_format(self, format):
        format.model_type.__marsha_format__ = format



from dateutil import parser
from datetime import datetime
from .base_format import BaseFormat


# TODO: Break this out into a module named marsha_dateutil
class ISODateString(BaseFormat):
    python_type = datetime
    formatted_type = str

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def load(self, data: formatted_type) -> python_type:
        return parser.parse(data, **self.kwargs).date()

    def dump(self, data: python_type, parent=None) -> formatted_type:
        return data.isoformat()



from distutils.core import setup

setup(
    name='marsha',
    version='0.2.0',
    packages=['marsha', 'marsha.formats'],
    license='MIT',
    url='https://gitlab.com/deckar01/marsha',
    author='Jared Deckard',
    author_email='jared@shademaps.com',
    long_description=open('README.md').read(),
)



from .raw_format import RawFormat


class String(RawFormat):
    model_type = str
    formatted_type = str

    @classmethod
    def min(self, length):
        self.min_length = length

        def min_constraint(value):
            if len(value) < self.min_length:
                error = 'Expected a minimum length of {}, but found {}.'
                return error.format(self.min_length, len(value))
        self.validators.append(min_constraint)

    @classmethod
    def max(self, length):
        self.max_length = length

        def max_constraint(value):
            if len(value) > self.max_length:
                error = 'Expected a maximum length of {}, but found {}.'
                return error.format(self.max_length, len(value))
        self.validators.append(max_constraint)



from .object import Object


class Dict(Object):
    model_type = dict
    formatted_type = dict
    schema_class = object

    @classmethod
    def validate_content(self, data):
        errors = {}
        for key, format in self.child_formats:
            error = format.validate_content(data[key])
            if error:
                errors[key] = error
        return errors or super(Object, self).validate_content(data)

    @classmethod
    def dump(self, data: model_type) -> formatted_type:
        formatted_data = self.get_extra(data)
        for key, format in self.child_formats:
            data_key = format.formatted_name
            formatted_data[data_key] = format.dump(data[key])
        return formatted_data

    @classmethod
    def compile_dump_item(self, scope, key):
        return scope + '["' + key + '"]'
